/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 IITP RAS
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Pavel Boyko <boyko@iitp.ru>
 */

/*
 * Classical hidden terminal problem and its RTS/CTS solution.
 *
 * Topology: [node 0] <-- -50 dB --> [node 1] <-- -50 dB --> [node 2]
 *
 * This example illustrates the use of
 *  - Wifi in ad-hoc mode
 *  - Matrix propagation loss model
 *  - Use of OnOffApplication to generate CBR stream
 *  - IP flow monitor
 */
#include <iostream>
#include <fstream>
#include <string>

#include "ns3/core-module.h"
#include "ns3/propagation-module.h"
#include "ns3/network-module.h"
#include "ns3/applications-module.h"
#include "ns3/mobility-module.h"
#include "ns3/internet-module.h"
#include "ns3/flow-monitor-module.h"
#include "ns3/wifi-module.h"
#include "ns3/ipv4-address-helper.h"
#include "ns3/ipv4-static-routing-helper.h"
#include "ns3/netanim-module.h"
#include "ns3/nstime.h"
#include "ns3/buffer.h"
#include <dirent.h>
#include <sys/stat.h>
#include "ns3/traffic-control-module.h"


#define PROG_DIR "./scratch/UDP_adhoc/"

//Prameter Setting
#define FIRST_LOAD  0.1
#define LAST_LOAD   5.0
#define DIF_LOAD 0.1



#define SIM_END_TIME  50
#define LOAD_FLOW2 0.0001
#define CW_INTERVAL     0.1
#define TH_INTERVAL 	0.1

using namespace ns3;

NS_LOG_COMPONENT_DEFINE ("TcpString");

double off =0; // bytes
double off_flow2 =0; // bytes
double off_load=0;
double thr_flow1=0;
double thr_flow2=0;

int HOP=2;
//#define NODE HOP+1
std::string fpresult="non";
int PAY=472;

uint32_t IP_paysize = (uint32_t) PAY;
double rate_MAC_PAY_to_IP_PAY = IP_paysize/(28.0+IP_paysize);


Time delay_flow1;
Time delay_flow2;
Ptr<OutputStreamWrapper> cWndStream;
Ptr<OutputStreamWrapper> rttStream;
uint32_t cWndValue;

bool firstRtt = true;

uint32_t oldTotalBytes=0;
uint32_t newTotalBytes;

#include <sstream>      // std::ostringstream
template <typename T> std::string tostr(const T& t)
{
    std::ostringstream os; os<<t; return os.str();
}

void SetStaticRoute(Ptr<Node> n, const char* destination, const char* nextHop, uint32_t interface) {
    Ipv4StaticRoutingHelper staticRouting;
    Ptr<Ipv4> ipv4 = n->GetObject<Ipv4> ();
    Ptr<Ipv4StaticRouting> a = staticRouting.GetStaticRouting (ipv4); a->AddHostRouteTo (Ipv4Address (destination), Ipv4Address (nextHop), interface);
}


void CwndTracer (Ptr<OutputStreamWrapper>stream, uint32_t oldcwnd, uint32_t newcwnd)
{
    *stream->GetStream() << Simulator::Now ().GetSeconds ()
    << "\t " << newcwnd << std::endl;
}

void
TraceThroughput (Ptr<Application> app, Ptr<OutputStreamWrapper> stream)
{
    Ptr <PacketSink> pktSink = DynamicCast <PacketSink> (app);
    newTotalBytes = pktSink->GetTotalRx ();
    *stream->GetStream() << Simulator::Now ().GetSeconds ()
    << " \t " << (newTotalBytes - oldTotalBytes)*8.0 / TH_INTERVAL / 1000 / 1000 /rate_MAC_PAY_to_IP_PAY<< std::endl;
    oldTotalBytes = newTotalBytes;
    Simulator::Schedule (Seconds (TH_INTERVAL), &TraceThroughput, app, stream);
}


void MyEventHandller (Ptr<Application> app, Ptr<OutputStreamWrapper> st)
{
    Ptr<Socket> src_socket  = app->GetObject<OnOffApplication>()->GetSocket();
    src_socket->TraceConnectWithoutContext("CongestionWindow",
                                           MakeBoundCallback(&CwndTracer, st));
}

void MyEventWrapper (Ptr<Application> app, Ptr<OutputStreamWrapper> st)
{
    Simulator::Schedule (Seconds (CW_INTERVAL), &MyEventHandller, app, st);
}

static void
RttTracer (Ptr<OutputStreamWrapper>stream, Time oldval, Time newval)
{
    *stream->GetStream() << Simulator::Now ().GetSeconds ()
    << "\t " << newval.GetSeconds () << std::endl;
}


void MyEventHandller_rtt (Ptr<Application> app, Ptr<OutputStreamWrapper> st)
{
    Ptr<Socket> src_socket  = app->GetObject<OnOffApplication>()->GetSocket();
    src_socket->TraceConnectWithoutContext("RTT",
                                           MakeBoundCallback(&RttTracer, st));
}

void MyEventWrapper_rtt (Ptr<Application> app, Ptr<OutputStreamWrapper> st)
{
    Simulator::Schedule (Seconds (CW_INTERVAL), &MyEventHandller_rtt, app, st);
}


void experiment (bool enableCtsRts,int argc, char **argv)
{
    int j;
    int NODE=HOP+1;
    
    double off2 = off * rate_MAC_PAY_to_IP_PAY;
    double CSThreshold = -82.0;
    std::string offeredload = std::to_string(2*off2) + "Mbps";
    std::cout<<off<<"\n";
    
    // 0. Enable or disable CTS/RTS
    UintegerValue ctsThr = (enableCtsRts ? UintegerValue (100) : UintegerValue (2200));
    Config::SetDefault ("ns3::WifiRemoteStationManager::RtsCtsThreshold", ctsThr); //RTS/CTSモード
    
    // 1. Create 3 nodes
    NodeContainer nodes;
    nodes.Create (NODE);
    
    // 2. Place nodes somehow, this is required by every wireless simulation
    for (j = 0; j < NODE; ++j)
    {
        nodes.Get (j)->AggregateObject (CreateObject<ConstantPositionMobilityModel> ());
    }
    
    
    // 5. Install wireless devices
    // PHY Layer setting
    WifiHelper wifi;
    wifi.SetStandard (WIFI_PHY_STANDARD_80211a);
    wifi.SetRemoteStationManager ("ns3::ConstantRateWifiManager",
                                  "DataMode",StringValue ("OfdmRate18Mbps"),
                                  "ControlMode",StringValue ("OfdmRate12Mbps"));
    YansWifiChannelHelper wifiChannel;
    wifiChannel.SetPropagationDelay ("ns3::ConstantSpeedPropagationDelayModel");
    wifiChannel.AddPropagationLoss ("ns3::LogDistancePropagationLossModel");
    YansWifiPhyHelper wifiPhy =  YansWifiPhyHelper::Default ();
    wifiPhy.SetPcapDataLinkType (YansWifiPhyHelper::DLT_IEEE802_11_RADIO);
    wifiPhy.SetChannel (wifiChannel.Create());
    wifiPhy.Set ("EnergyDetectionThreshold", DoubleValue (-90) );
    wifiPhy.Set ("CcaMode1Threshold", DoubleValue (CSThreshold) );
    

    // MAC Layer setting
    NqosWifiMacHelper wifiMac = NqosWifiMacHelper::Default ();
    wifiMac.SetType ("ns3::AdhocWifiMac"); // use ad-hoc MAC
    NetDeviceContainer devices = wifi.Install (wifiPhy,wifiMac,nodes);
    
    // Position string-topology network
    MobilityHelper mobility;
    Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();
    for (j=0;j<NODE+1;j++){
        positionAlloc->Add (Vector (j*45, 200.0, 0.0));
    }
    mobility.SetPositionAllocator (positionAlloc);
    mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
    mobility.Install (nodes);
    

    // 6. Install TCP/IP stack & assign IP addresses
    InternetStackHelper internet;
    internet.Install (nodes);
    Ipv4AddressHelper ipv4;
    ipv4.SetBase ("192.168.1.0", "255.255.255.0");
    ipv4.Assign (devices);
    for (j=0; j<NODE; j++) {
        Ptr<Ipv4> ipv4node0 = nodes.Get(j)->GetObject<Ipv4>();
    }
    char ip_node[NODE][13];
    for (j=0; j<NODE; j++) {
        sprintf(ip_node[j],"192.168.1.%d",j+1);
        std::cout << "IP of NODE" <<j<<"::"<<"192.168.1."<<j<<"\n";
    }
    
    // Static rounting for string topology
    for (j=0; j<NODE; j++) {
        if (j==0) {
            SetStaticRoute(nodes.Get(j),ip_node[NODE-1],ip_node[j+1], 1);
        }else if(j==NODE-1){
            SetStaticRoute(nodes.Get(j),ip_node[0],ip_node[j-1], 1);
        }else{
            SetStaticRoute(nodes.Get(j),ip_node[NODE-1],ip_node[j+1], 1);
            SetStaticRoute(nodes.Get(j),ip_node[0],ip_node[j-1], 1);
        }
    }
    
    // 7. Install applications: two CBR streams each saturating the channel
    ApplicationContainer cbrApps;
    uint16_t cbrPort1 = 12345;
    OnOffHelper flow1 ("ns3::UdpSocketFactory", InetSocketAddress (Ipv4Address (ip_node[NODE-1]), cbrPort1));
    flow1.SetAttribute ("PacketSize", UintegerValue (IP_paysize));
    double mean=((IP_paysize+28)*8)/(2*off2*1000000);
    std::string set_exp_mean ="ns3::ExponentialRandomVariable[Mean="+std::to_string(mean)+"]";
    flow1.SetAttribute ("OnTime",  StringValue (set_exp_mean));
    flow1.SetAttribute ("OffTime",  StringValue (set_exp_mean));
    flow1.SetAttribute ("DataRate", StringValue (offeredload));
    flow1.SetAttribute ("StartTime", TimeValue (Seconds (1.000000)));
    cbrApps.Add (flow1.Install (nodes.Get (0)));
    
    ApplicationContainer cbrApps2;
    uint16_t cbrPort2 = 40000;
    OnOffHelper flow2 ("ns3::UdpSocketFactory", InetSocketAddress (Ipv4Address (ip_node[0]), cbrPort2));
    flow2.SetAttribute ("PacketSize", UintegerValue (IP_paysize));
    flow2.SetAttribute ("OnTime",  StringValue (set_exp_mean));
    flow2.SetAttribute ("OffTime",  StringValue (set_exp_mean));
    flow2.SetAttribute ("DataRate", StringValue ("0.01Mbps"));
    flow2.SetAttribute ("StartTime", TimeValue (Seconds (1.000000)));
    cbrApps.Add (flow2.Install (nodes.Get (NODE-1)));
    
    
    Address sinkAddress (InetSocketAddress (ip_node[NODE-1],cbrPort1));
    PacketSinkHelper sinkHelper ("ns3::UdpSocketFactory", sinkAddress);
    ApplicationContainer sink_cbr = sinkHelper.Install (nodes.Get(NODE-1));
    
    
    Address sinkAddress2 (InetSocketAddress (ip_node[0],cbrPort2));
    PacketSinkHelper sinkHelper2 ("ns3::UdpSocketFactory", sinkAddress2);
    ApplicationContainer sink_cbr2 = sinkHelper2.Install (nodes.Get(0));
    
    uint16_t  echoPort = 9;
    UdpEchoClientHelper echoClientHelper (Ipv4Address (ip_node[NODE-1]), echoPort);
    echoClientHelper.SetAttribute ("MaxPackets", UintegerValue (1));
    echoClientHelper.SetAttribute ("Interval", TimeValue (Seconds (0.1)));
    echoClientHelper.SetAttribute ("PacketSize", UintegerValue (10));
    ApplicationContainer pingApps;
    
    // again using different start times to workaround Bug 388 and Bug 912
    echoClientHelper.SetAttribute ("StartTime", TimeValue (Seconds (0.001)));
    pingApps.Add (echoClientHelper.Install (nodes.Get (0)));
    
    // 8. Install FlowMonitor on all nodes
    FlowMonitorHelper flowmon;
    Ptr<FlowMonitor> monitor = flowmon.InstallAll ();
    
    
    // 9. Run simulation for 10 seconds
    Simulator::Stop (Seconds (SIM_END_TIME));
    Simulator::Run ();
    
    // 10. Print per flow statistics
    monitor->CheckForLostPackets ();
    Ptr<Ipv4FlowClassifier> classifier = DynamicCast<Ipv4FlowClassifier> (flowmon.GetClassifier ());
    FlowMonitor::FlowStatsContainer stats = monitor->GetFlowStats ();
    
    
    for (std::map<FlowId, FlowMonitor::FlowStats>::const_iterator i = stats.begin (); i != stats.end (); ++i)
    {
        
        if (i->first > 1)
        {
            Ipv4FlowClassifier::FiveTuple t = classifier->FindFlow (i->first);
            std::cout << " Flow " << i->first - 1 << " (" << t.sourceAddress << " -> " << t.destinationAddress << ")\n";
            std::cout << "  Delay sum: " << i->second.delaySum/i->second.rxPackets/1000 << "\n";
            std::cout << "  Jitter sum: " << i->second.jitterSum/i->second.rxPackets/1000 << "\n";
            std::cout << "  Tx Packets: " << i->second.txPackets << "\n";
            std::cout << "  Tx Bytes:   " << i->second.txBytes << "\n";
            std::cout << "  Paysize:   " << i->second.txBytes/i->second.txPackets  << "\n";
            int payload_size=i->second.txBytes/i->second.txPackets;
            std::cout << "  Time last: " << i->second.timeLastTxPacket/1000000000 << "\n";
            std::cout << "  TxOffered:  " << i->second.txPackets * (payload_size) * 8.0 / (SIM_END_TIME-1) / 1000 / 1000  << " Mbps\n";
            std::cout << "  Rx Packets: " << i->second.rxPackets << "\n";
            std::cout << "  Rx Bytes:   " << i->second.rxBytes << "\n";
            std::cout << "  Throughput: " << i->second.rxPackets * (payload_size) * 8.0 /(SIM_END_TIME-1) / 1000 / 1000  << " Mbps\n";
            if (i->first - 1==1) {
                thr_flow1 =i->second.rxPackets * (payload_size) * 8.0 / (SIM_END_TIME-1) / 1000 / 1000;
                off_load =i->second.txPackets * (payload_size) * 8.0 / (SIM_END_TIME-1) / 1000 / 1000;
                delay_flow1 = i->second.delaySum/i->second.rxPackets/1000;
            }else{
                thr_flow2 =i->second.rxPackets * (payload_size) * 8.0 / (SIM_END_TIME-1)/ 1000 / 1000;
                delay_flow2 = i->second.delaySum/i->second.rxPackets/1000;
            }
        }
    }
    // 11. Cleanup
    Simulator::Destroy ();
}

int main (int argc, char **argv)
{
    
    CommandLine cmd;
    cmd.AddValue ("HOP", "Hop number", HOP);
    cmd.AddValue ("fp_result", "Name of result Forder",fpresult);
    cmd.AddValue ("paysize", "Name of result Forder",PAY);
    cmd.Parse (argc, argv);
    std::cout<<HOP<<"\n";
    std::cout<<fpresult<<"\n";
    std::string fp_thr=std::string(PROG_DIR)+fpresult+"/throughput.txt";
    std::ofstream thr(fp_thr,std::ios::out);
    std::string fp_delay=std::string(PROG_DIR) + fpresult+ "/delay.txt";
    std::ofstream delay(fp_delay,std::ios::out);
    thr <<"off1 \t off1+off1 \t thr_1 \t thr_2 \t thr_1+thr_2\n";
    for (off=FIRST_LOAD; off<=LAST_LOAD; off+=DIF_LOAD) {
        experiment (false,argc,argv);
        thr << off <<"\t"<< off_load <<"\t"<< thr_flow1 << "\t" <<  thr_flow2 << "\t"<< thr_flow1+thr_flow2 << "\n";
        delay << off << "\t" <<delay_flow1<< "\t" <<delay_flow2 << "\n";
        std::cout << "------------------------------------------------\n";
        
    }
    return 0;
}
